/*
 * location.c - Source code location of language elements
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */

/*
 * KNOWN BUG: outputs tunnels once per selector @@@
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "util.h"
#include "error.h"
#include "tree.h"
#include "device.h"
#include "police.h"
#include "filter.h"
#include "qdisc.h"
#include "location.h"


char *file_name = NULL;
int lineno = 1;

static int file_name_ref = 0; /* reference count for "file_name" */


/* ------------------------------- Registry -------------------------------- */


void input_file(const char *name)
{
    if (file_name && !file_name_ref) free(file_name);
    file_name = stralloc(name);
    file_name_ref = 0;
}


LOCATION current_location(void)
{
    LOCATION loc;

    loc.file = file_name ? file_name : "(stdin)";
    loc.tag = NULL;
    loc.line = lineno;
    file_name_ref++;
    return loc;
}


int print_location(FILE *file,LOCATION loc)
{
    return fprintf(file,"%s:%d",loc.file,loc.line);
}


int print_current_location(FILE *file)
{
    return fprintf(file,"%s:%d",file_name ? file_name : "(stdin)",lineno);
}


/* --------------------------- Dump location map --------------------------- */


static void do_map(FILE *file,LOCATION loc)
{
    fprintf(file," %s %s:%d\n",loc.tag ? loc.tag : "-",loc.file,loc.line);
}


static void id_qdisc(FILE *file,const QDISC *qdisc)
{
    fprintf(file,"%s:",qdisc->parent.device->name);
    if (qdisc->dsc == &ingress_dsc) fprintf(file,"ingress");
    else fprintf(file,"%u",(unsigned) qdisc->number);
}


static void id_class(FILE *file,const CLASS *class)
{
     id_qdisc(file,class->parent.qdisc);
     fprintf(file,":%u",(unsigned) class->number);
}


static void id_filter(FILE *file,const FILTER *filter)
{
    if (filter->parent.class) id_class(file,filter->parent.class);
    else {
	id_qdisc(file,filter->parent.qdisc);
	fprintf(file,":");
    }
    fprintf(file,":%u",(unsigned) filter->number);
}


static void map_elements(FILE *file,const ELEMENT *elements)
{
    const ELEMENT *element;
    int n = 0;

    for (element = elements; element; element = element->next) {
	fprintf(file,"element ");
	id_filter(file,element->parent.filter);
	fprintf(file,":%d",n);
	do_map(file,element->location);
	if (element->parent.class == &class_is_tunnel) {
	    const ELEMENT *scan;

	    for (scan = elements; scan != element; scan = scan->next)
		if (scan->parent.class == &class_is_tunnel &&
		  scan->parent.tunnel == element->parent.tunnel)
		    break;
	    if (scan == element) {
		fprintf(file,"tunnel ");
		id_filter(file,element->parent.filter);
		fprintf(file,":%u",(unsigned) element->parent.tunnel->number);
		do_map(file,element->parent.tunnel->location);
	    }
	}
	n++;
    }
}


static void map_filters(FILE *file,const FILTER *filters)
{
    const FILTER *filter;

    for (filter = filters; filter; filter = filter->next) {
	fprintf(file,"filter ");
	id_filter(file,filter);
	do_map(file,filter->location);
	map_elements(file,filter->elements);
    }
}


static void map_qdisc(FILE *file,const QDISC *qdisc);


static void map_classes(FILE *file,const CLASS *classes)
{
    const CLASS *class;

    for (class = classes; class; class = class->next) {
	fprintf(file,"class ");
	id_class(file,class);
	do_map(file,class->location);
	map_classes(file,class->child);
	map_qdisc(file,class->qdisc);
	map_filters(file,class->filters);
    }
}


static void map_qdisc(FILE *file,const QDISC *qdisc)
{
    if (!qdisc) return;
    fprintf(file,"qdisc ");
    id_qdisc(file,qdisc);
    do_map(file,qdisc->location);
    map_classes(file,qdisc->classes);
    map_filters(file,qdisc->filters);
}


static void map_devices(FILE *file)
{
    const DEVICE *device;

    for (device = devices; device; device = device->next) {
	fprintf(file,"device %s",device->name);
	do_map(file,device->location);
	map_qdisc(file,device->ingress);
	map_qdisc(file,device->egress);
    }
}


static void map_policers(FILE *file)
{
    const POLICE *p;

    for (p = policers; p; p = p->next)
	if (p->used) {
	    fprintf(file,"police %u",(unsigned) p->number);
	    do_map(file,p->location);
	}
}


void write_location_map(const char *file_name)
{
    FILE *file;

    fflush(stdout);
    fflush(stderr);
    if (!strcmp(file_name,"stderr")) file = stderr;
    else {
	file = fopen(file_name,"w");
	if (!file) {
	    perror(file_name);
	    exit(1);
	}
    }
    map_devices(file);
    map_policers(file);
    fflush(file);
    if (ferror(file)) errorf("error writing to \"%s\"",file_name);
    if (strcmp(file_name,"stderr"))
	if (fclose(file) == EOF) {
	    perror(file_name);
	    exit(1);
	}
}
